@@ -26,6 +26,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <glade/glade-xml.h>
+#include <glade/glade-init.h>
 #include <glade/glade-build.h>
 #include <glade/glade-private.h>
 #include <gmodule.h>
@@ -34,6 +35,7 @@
 #include <gtk/gtkwidget.h>
 #include <gtk/gtkcontainer.h>
 #include <gtk/gtklabel.h>
+#include <atk/atk.h>
 
 #ifdef ENABLE_NLS
 #  include <libintl.h>
@@ -98,6 +100,9 @@ glade_xml_class_init (GladeXMLClass *class)
     glade_xml_tree_id = g_quark_from_static_string(glade_xml_tree_key);
     glade_xml_name_id = g_quark_from_static_string(glade_xml_name_key);
     glade_xml_tooltips_id = g_quark_from_static_string(glade_xml_tooltips_key);
+
+    /* make sure libglade is initialised */
+    glade_init();
 }
 
 static void
@@ -112,6 +117,8 @@ glade_xml_init (GladeXML *self)
     priv->tree = NULL;
     priv->tooltips = gtk_tooltips_new();
     gtk_tooltips_enable(priv->tooltips);
+    gtk_object_ref(GTK_OBJECT(priv->tooltips));
+    gtk_object_sink(GTK_OBJECT(priv->tooltips));
     priv->name_hash = g_hash_table_new(g_str_hash, g_str_equal);
     priv->signals = g_hash_table_new(g_str_hash, g_str_equal);
     priv->radio_groups = g_hash_table_new (g_str_hash, g_str_equal);
@@ -680,9 +687,10 @@ glade_xml_handle_widget_prop(GladeXML *self, GtkWidget *widget,
     } else {
 	GladeDeferredProperty *dprop = g_new(GladeDeferredProperty, 1);
 
-	dprop->target = G_OBJECT(widget);
-	dprop->prop_name = prop_name;
-	dprop->prop_value = value_name;
+	dprop->target_name = value_name;
+	dprop->type = DEFERRED_PROP;
+	dprop->d.prop.object = G_OBJECT(widget);
+	dprop->d.prop.prop_name = prop_name;
 
 	self->priv->deferred_props = g_list_prepend(self->priv->deferred_props,
 						    dprop);
@@ -772,6 +780,106 @@ glade_xml_add_signals(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
     }
 }
 
+/* this is a private function */
+static void
+glade_xml_add_atk_actions(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
+{
+    gint i;
+    gint n;
+    AtkObject *atko = gtk_widget_get_accessible (w);
+    AtkAction *action;
+
+    if (info->n_atk_actions) {
+        if (!ATK_IS_ACTION (atko)) {
+            g_warning("widgets of type %s don't have actions, but one is specified",
+	  	  G_OBJECT_TYPE_NAME (w));
+	    return;
+        }
+
+        action = ATK_ACTION (atko);
+        n = atk_action_get_n_actions (action);    
+    
+        for (i = 0; i < info->n_atk_actions; i++) {
+    	    GladeAtkActionInfo *action_info = &info->atk_actions[i];
+	    int j;
+	    for (j=0; j<n; ++j) {
+	        if (strcmp (atk_action_get_name (action, j),
+			    action_info->action_name) == 0) {
+	            break;
+	        }
+	    }
+	    if (j < n) {
+	        atk_action_set_description (ATK_ACTION (atko),
+					    j,
+					    action_info->description);	
+	    } else {
+		/* don't show a warning -- the action names might not
+                 * be registered if libgail hasn't been loaded */
+
+		/* g_warning("could not find action named %s", action_info->action_name); */
+	    }
+        }
+    }	
+}
+
+/* helper function for adding relations */
+static void
+add_relation(AtkRelationSet *relations, AtkRelationType relation_type,
+	     AtkObject *target_accessible)
+{
+    AtkRelation *relation;
+
+    relation = atk_relation_set_get_relation_by_type (relations,
+						      relation_type);
+    if (relation) {
+	/* add new target accessible to relation */
+	GPtrArray* target_array = atk_relation_get_target (relation);
+
+	g_ptr_array_remove (target_array, target_accessible);
+	g_ptr_array_add (target_array, target_accessible);
+    } else {
+	/* the relation hasn't been created yet ... */
+	atk_relation_set_add (relations,
+			      atk_relation_new (&target_accessible, 1,
+						relation_type));
+    }
+}
+
+/* this is a private function */
+static void
+glade_xml_add_atk_relations(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
+{
+    gint i;
+    AtkObject *atko = gtk_widget_get_accessible (w);
+    AtkRelationSet *relations = atk_object_ref_relation_set (atko);
+    
+    for (i = 0; i < info->n_relations; i++) {
+	GladeAtkRelationInfo *rinfo = &info->relations[i];
+	GtkWidget *target_widget = glade_xml_get_widget (xml, rinfo->target);
+	AtkRelationType relation_type;
+
+	relation_type = atk_relation_type_for_name (rinfo->type);
+	if (target_widget) {
+	    AtkObject *target_accessible;
+
+	    target_accessible = gtk_widget_get_accessible (target_widget);
+
+	    add_relation(relations, relation_type, target_accessible);
+	} else {
+	    GladeDeferredProperty *dprop = g_new(GladeDeferredProperty, 1);
+
+	    dprop->target_name = rinfo->target;
+	    dprop->type = DEFERRED_REL;
+	    dprop->d.rel.relation_set = g_object_ref(relations);
+	    dprop->d.rel.relation_type = relation_type;
+
+	    xml->priv->deferred_props =
+		g_list_prepend(xml->priv->deferred_props, dprop);
+	}
+    }
+    g_object_unref (relations);
+}
+
 static void
 glade_xml_add_accels(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
 {
@@ -789,6 +897,43 @@ glade_xml_add_accels(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
     }
 }
 
+static void
+glade_xml_add_accessibility_info(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
+{
+    gint i;
+    AtkObject *atko = gtk_widget_get_accessible (w);
+    
+    for (i = 0; i < info->n_atk_props; i++) {
+	GParamSpec *pspec;
+	GValue value = { 0 };
+
+	pspec = g_object_class_find_property(g_type_class_ref(ATK_TYPE_OBJECT),
+					     info->atk_props[i].name);
+	if (!pspec) {
+	    g_warning("unknown property `%s' for class `%s'",
+		      info->atk_props[i].name, g_type_name(ATK_TYPE_OBJECT));
+	    continue;
+	} else if (glade_xml_set_value_from_string (xml, pspec, info->atk_props[i].value, &value)) {
+	    g_object_set_property (G_OBJECT (atko), info->atk_props[i].name, &value);
+	    g_value_unset (&value);
+        }
+	
+	debug(g_message("Adding accessibility property %s:%s",
+			info->atk_props[i].name,
+			info->atk_props[i].value));
+    }
+
+    /* Hang on, we're not done yet ;-) */
+    glade_xml_add_atk_actions(xml, w, info);
+    glade_xml_add_atk_relations(xml, w, info);
+    /*
+       hmm, might have to defer some relations,
+       if the widgets they refer to haven't
+       been constructed yet (will look at the other prop-as-widget
+       stuff for ideas)
+    */
+}
+
 static void
 glade_xml_destroy_signals(char *key, GList *signal_datas)
 {
@@ -811,7 +956,6 @@ static void
 remove_data_func(gpointer key, gpointer value, gpointer user_data)
 {
     GObject *object = value;
-    GladeXML *xml = user_data;
 
     g_object_set_qdata(object, glade_xml_tree_id, NULL);
     g_object_set_qdata(object, glade_xml_name_id, NULL);
@@ -975,32 +1119,49 @@ guint
 glade_flags_from_string (GType type, const char *string)
 {
     GFlagsClass *fclass;
-    GFlagsValue *fv;
     gchar *endptr;
-    guint ret = 0;
+    guint i, j, ret = 0;
+    char *flagstr;
 
     ret = strtoul(string, &endptr, 0);
     if (endptr != string) /* parsed a number */
 	return ret;
 
-    ret = 0;
     fclass = g_type_class_ref(type);
 
-    while ((endptr = strchr(string, ' '))) {
-	char *str = g_strndup(string, endptr-string);
 
-	fv = g_flags_get_value_by_name(fclass, str);
-	if (!fv) fv = g_flags_get_value_by_nick(fclass, str);
-	if (fv)  ret |= fv->value;
+    flagstr = g_strdup (string);
+    for (ret = i = j = 0; ; i++) {
+	gboolean eos;
 
-	g_free(str);
-	string = endptr;
-	while (string[0] == ' ' || string[0] == '|')
-	    string++;
+	eos = flagstr [i] == '\0';
+	
+	if (eos || flagstr [i] == '|') {
+	    GFlagsValue *fv;
+	    const char  *flag;
+
+	    flag = &flagstr [j];
+
+	    if (!eos) {
+		flagstr [i++] = '\0';
+		j = i;
+	    }
+
+	    fv = g_flags_get_value_by_name (fclass, flag);
+
+	    if (!fv)
+		fv = g_flags_get_value_by_nick (fclass, flag);
+
+	    if (fv)
+		ret |= fv->value;
+
+	    if (eos)
+		break;
+	} else
+	    i++;
     }
-    fv = g_flags_get_value_by_name(fclass, string);
-    if (!fv) fv = g_flags_get_value_by_nick(fclass, string);
-    if (fv)  ret |= fv->value;
+    
+    g_free (flagstr);
 
     g_type_class_unref(fclass);
 
@@ -1043,36 +1204,205 @@ glade_xml_build_interface(GladeXML *self, GladeInterface *iface,
 
 /* below are functions from glade-build.h */
 
-static GHashTable *widget_table = NULL;
+static GQuark glade_build_data_id = 0;
+static const gchar *glade_build_data_key = "libglade::build-data";
+typedef struct _GladeWidgetBuildData GladeWidgetBuildData;
+struct _GladeWidgetBuildData {
+    GladeNewFunc new;
+    GladeBuildChildrenFunc build_children;
+    GladeFindInternalChildFunc find_internal_child;
+};
 
 /**
- * glade_register_widgets:
- * @widgets: a NULL terminated array of GladeWidgetBuildData structures.
+ * glade_register_widget:
+ * @type: the GType of the widget.
+ * @new: the function used to construct instances of the widget.
+ * @build_children: function used to construct children (or NULL).
+ * @find_internal_child: function to find internal children (or NULL).
  *
- * This function is used to register new sets of widget building functions.
- * each member of @widgets contains the widget name, a function to build
- * a widget of that type, and optionally a function to build the children
- * of this widget.  The child building routine would call
- * glade_xml_build_widget on each child node to create the child before
- * packing it.
+ * This function is used to register new construction functions for a
+ * widget type.  The child building routine would call
+ * glade_xml_build_widget on each child node to create the child
+ * before packing it.
  *
  * This function is mainly useful for addon widget modules for libglade
  * (it would get called from the glade_init_module function).
  */
 void
-glade_register_widgets(const GladeWidgetBuildData *widgets)
+glade_register_widget(GType type,
+		      GladeNewFunc new,
+		      GladeBuildChildrenFunc build_children,
+		      GladeFindInternalChildFunc find_internal_child)
 {
-    int i = 0;
+    GladeWidgetBuildData *data;
 
-    if (!widget_table)
-	widget_table = g_hash_table_new(g_str_hash, g_str_equal);
-    while (widgets[i].name != NULL) {
-	g_hash_table_insert(widget_table, widgets[i].name,
-			    (gpointer)&widgets[i]);
-	i++;
-    }
+    g_return_if_fail(g_type_is_a(type, GTK_TYPE_WIDGET));
+
+    if (glade_build_data_id == 0)
+	glade_build_data_id = g_quark_from_static_string(glade_build_data_key);
+
+    if (!new) new = glade_standard_build_widget;
+
+    data = g_new(GladeWidgetBuildData, 1);
+
+    data->new = new;
+    data->build_children = build_children;
+    data->find_internal_child = find_internal_child;
+
+    g_type_set_qdata(type, glade_build_data_id, data);
 }
 
+/* helper function for getting the build data for a type */
+static const GladeWidgetBuildData *
+get_build_data(GType type)
+{
+    static const GladeWidgetBuildData widget_build_data = {
+	glade_standard_build_widget,
+	NULL,
+	NULL
+    };
+    static const GladeWidgetBuildData container_build_data = {
+	glade_standard_build_widget,
+	glade_standard_build_children,
+	NULL
+    };
+    const GladeWidgetBuildData *build_data;
+
+    if (glade_build_data_id == 0)
+	glade_build_data_id = g_quark_from_static_string(glade_build_data_key);
+
+    build_data = g_type_get_qdata(type, glade_build_data_id);
+    
+    if (build_data)
+	return build_data;
+    if (g_type_is_a(type, GTK_TYPE_CONTAINER))
+	return &container_build_data;
+    else
+	return &widget_build_data;
+}
+
+
+/* used to store a GArray of custom property handlers */
+static GQuark       glade_custom_props_id = 0;
+static const gchar *glade_custom_props_key = "libglade::custom-props";
+/* used to store a cache of the property handlers (ie. this type's
+ * properties + parent properties) */
+static GQuark       glade_custom_props_cache_id = 0;
+static const gchar *glade_custom_props_cache_key = "libglade::custom-props-cache";
+typedef struct _CustomPropInfo CustomPropInfo;
+struct _CustomPropInfo {
+    GQuark name_quark;
+    GladeApplyCustomPropFunc apply_prop;
+};
+
+/* utility function for invalidating the cached property key */
+static void
+invalidate_custom_prop_cache(GType type)
+{
+    gpointer data;
+    GType *children;
+    guint i, n_children = 0;
+
+    data = g_type_get_qdata(type, glade_custom_props_cache_id);
+    /* if there is no cached custom prop data, then no children will
+     * have cached data. */
+    if (!data)
+	return;
+
+    g_type_set_qdata(type, glade_custom_props_cache_id, NULL);
+    g_free(data);
+    children = g_type_children(type, &n_children);
+    for (i = 0; i < n_children; i++)
+	invalidate_custom_prop_cache(children[i]);
+    g_free(children);
+}
+
+void
+glade_register_custom_prop(GType type, const gchar *prop_name,
+			   GladeApplyCustomPropFunc apply_prop)
+{
+    GArray *array;
+    CustomPropInfo prop_info;
+    
+    /* set up quarks ... */
+    if (glade_custom_props_id == 0) {
+	glade_custom_props_id = g_quark_from_static_string(glade_custom_props_key);
+	glade_custom_props_cache_id = g_quark_from_static_string(glade_custom_props_cache_key);
+    }
+
+    /* get the GArray of prop info */
+    array = g_type_get_qdata(type, glade_custom_props_id);
+    if (!array) {
+	array = g_array_new(TRUE, FALSE, sizeof(CustomPropInfo));
+	g_type_set_qdata(type, glade_custom_props_id, array);
+    }
+
+    /* append the extra prop */
+    prop_info.name_quark = g_quark_from_string(prop_name);
+    prop_info.apply_prop = apply_prop;
+    g_array_append_val(array, prop_info);
+
+    /* invalidate cached info, if any */
+    invalidate_custom_prop_cache(type);
+}
+
+static CustomPropInfo *
+get_custom_prop_info(GType type)
+{
+    CustomPropInfo *prop_info, *parent_info = NULL;
+    GType parent;
+    GArray *array;
+    gint length;
+
+    /* if glade_register_custom_prop hasn't been called ... */
+    if (glade_custom_props_id == 0) {
+	return NULL;
+    }
+
+    /* has the prop info been calculated/cached already? */
+    prop_info = g_type_get_qdata(type, glade_custom_props_cache_id);
+    if (prop_info)
+	return prop_info;
+
+    array = g_type_get_qdata(type, glade_custom_props_id);
+    parent = g_type_parent(type);
+    prop_info = NULL;
+    length = 0;
+    if (parent == G_TYPE_INVALID) {/* hit the top */
+	if (array) {
+	    length = array->len;
+	    prop_info = g_memdup(array->data,
+				 sizeof(CustomPropInfo) * (length + 1));
+	}
+    } else { /* merge parent info with our info */
+	gint count = 0;
+
+	parent_info = get_custom_prop_info(parent);
+	if (parent_info)
+	    while (parent_info[count].name_quark != 0)
+		count++;
+	length = count;
+	if (array) length += array->len;
+	prop_info = g_new(CustomPropInfo, length + 1);
+	/* concatenate info */
+	if (count > 0)
+	    memcpy(prop_info, parent_info,
+		   sizeof(CustomPropInfo) * count);
+	if (array)
+	    memcpy(&prop_info[count], array->data,
+		   sizeof(CustomPropInfo) * array->len);
+    }
+    /* make sure it is null terminated */
+    if (prop_info != NULL) {
+	prop_info[length].name_quark = 0;
+	prop_info[length].apply_prop = NULL;
+    }
+
+    g_type_set_qdata(type, glade_custom_props_cache_id, prop_info);
+    return prop_info;
+}
+
+
 /**
  * glade_xml_set_value_from_string
  * @xml: the GladeXML object.
@@ -1201,6 +1531,13 @@ glade_xml_set_value_from_string (GladeXML *xml,
     return ret;
 }
 
+/* a small struct to hold info about a custom prop until we apply it */
+typedef struct _CustomPropData CustomPropData;
+struct _CustomPropData {
+    GladeApplyCustomPropFunc apply_prop;
+    GladeProperty *prop;
+};
+
 /**
  * glade_standard_build_widget
  * @xml: the GladeXML object.
@@ -1219,22 +1556,47 @@ glade_standard_build_widget(GladeXML *xml, GType widget_type,
 			    GladeWidgetInfo *info)
 {
     static GArray *props_array = NULL;
+    static GArray *custom_props_array = NULL;
     GObjectClass *oclass;
+    CustomPropInfo *custom_props;
     GtkWidget *widget;
     GList *deferred_props = NULL, *tmp;
     guint i;
 
-    if (!props_array)
+    if (!props_array) {
 	props_array = g_array_new(FALSE, FALSE, sizeof(GParameter));
+	custom_props_array = g_array_new(FALSE, FALSE, sizeof(CustomPropData));
+    }
 
     /* we ref the class here as a slight optimisation */
     oclass = g_type_class_ref(widget_type);
 
+    custom_props = get_custom_prop_info(widget_type);
+
     /* collect properties */
     for (i = 0; i < info->n_properties; i++) {
+	GQuark name_quark;
 	GParameter param = { NULL };
 	GParamSpec *pspec;
 
+	if (custom_props &&
+	    (name_quark = g_quark_try_string(info->properties[i].name)) != 0) {
+	    gint j = 0;
+
+	    while (custom_props[j].name_quark) {
+		if (custom_props[j].name_quark == name_quark) {
+		    CustomPropData prop_data;
+
+		    prop_data.apply_prop = custom_props[j].apply_prop;
+		    prop_data.prop = &info->properties[i];
+		    g_array_append_val(custom_props_array, prop_data);
+		    break;
+		}
+		j++;
+	    }
+	    if (custom_props[j].name_quark != 0) /* a prop was matched */
+		continue;
+	}
 	pspec = g_object_class_find_property(oclass, info->properties[i].name);
 	if (!pspec) {
 	    g_warning("unknown property `%s' for class `%s'",
@@ -1272,6 +1634,16 @@ glade_standard_build_widget(GladeXML *xml, GType widget_type,
 	g_value_unset(&g_array_index(props_array, GParameter, i).value);
     }
 
+    /* do custom props */
+    for (i = 0; i < custom_props_array->len; i++) {
+	CustomPropData *data;
+
+	data = &g_array_index(custom_props_array, CustomPropData, i);
+	if (data->apply_prop)
+	    (* data->apply_prop) (xml, widget, data->prop->name,
+				  data->prop->value);
+    }
+
     /* handle deferred properties */
     for (tmp = deferred_props; tmp; tmp = tmp->next) {
 	GladeProperty *prop = tmp->data;
@@ -1281,11 +1653,43 @@ glade_standard_build_widget(GladeXML *xml, GType widget_type,
     g_list_free(tmp);
 
     g_array_set_size(props_array, 0);
+    g_array_set_size(custom_props_array, 0);
     g_type_class_unref(oclass);
 
     return widget;
 }
 
+/**
+ * glade_xml_set_packing_property:
+ * @self: the GladeXML object.
+ * @parent: the container widget.
+ * @child: the contained child
+ * @name: the name of the property
+ * @value: it's stringified value
+ * 
+ * This sets the packing property on container @parent of widget
+ * @child with @name to @value
+ **/
+static void
+glade_xml_set_packing_property (GladeXML   *self,
+				GtkWidget  *parent, GtkWidget  *child,
+				const char *name,   const char *value)
+{
+    GValue gvalue = { 0 };
+    GParamSpec *pspec;
+
+    pspec = gtk_container_class_find_child_property(
+	G_OBJECT_GET_CLASS(parent), name);
+    if (!pspec)
+	g_warning("unknown child property `%s' for container `%s'",
+		  name, G_OBJECT_TYPE_NAME(parent));
+    else if (glade_xml_set_value_from_string(self, pspec, value, &gvalue)) {
+	gtk_container_child_set_property(GTK_CONTAINER(parent), child,
+					 name, &gvalue);
+	g_value_unset(&gvalue);
+    }
+}
+
 /**
  * glade_standard_build_children
  * @self: the GladeXML object.
@@ -1322,29 +1726,17 @@ glade_standard_build_children(GladeXML *self, GtkWidget *parent,
 
 	g_object_ref(G_OBJECT(child));
 	gtk_widget_freeze_child_notify(child);
+
 	gtk_container_add(GTK_CONTAINER(parent), child);
-	for (j = 0; j < info->children[i].n_properties; j++) {
-	    const gchar *name = info->children[i].properties[j].name;
-	    GValue value = { 0 };
-	    GParamSpec *pspec;
 
-	    pspec = gtk_container_class_find_child_property(
-		G_OBJECT_GET_CLASS(parent), name);
-	    if (!pspec) {
-		g_warning("unknown child property `%s' for container `%s'",
-			  name, G_OBJECT_TYPE_NAME(parent));
-		continue;
-	    }
+	for (j = 0; j < info->children[i].n_properties; j++)
+	    glade_xml_set_packing_property (
+		self, parent, child,
+		info->children[i].properties[j].name,
+		info->children[i].properties[j].value);
 
-	    if (glade_xml_set_value_from_string(self, pspec,
-			info->children[i].properties[j].value, &value)) {
-		gtk_container_child_set_property(GTK_CONTAINER(parent), child,
-						 name, &value);
-		g_value_unset(&value);
-	    }
-	}
 	gtk_widget_thaw_child_notify(child);
-	g_object_ref(G_OBJECT(child));
+	g_object_unref(G_OBJECT(child));
     }
     g_object_unref(G_OBJECT(parent));
 }
@@ -1420,30 +1812,27 @@ glade_standard_build_children(GladeXML *self, GtkWidget *parent,
 GtkWidget *
 glade_xml_build_widget(GladeXML *self, GladeWidgetInfo *info)
 {
-    GladeWidgetBuildData *data;
-    GType widget_type;
+    GType type;
     GtkWidget *ret;
 
-    debug(g_message("Widget class: %s", info->class));
+    debug(g_message("Widget class: %s\tname: %s", info->class, info->name));
     if (!strcmp(info->class, "Placeholder")) {
-	g_warning("placeholders exist in interface description");
+	static int warned = 0;
+	if (!warned++)
+	    g_warning("placeholders exist in interface description");
 	ret = gtk_label_new("[placeholder]");
 	gtk_widget_show(ret);
 	return ret;
     }
-    data = g_hash_table_lookup(widget_table, info->class);
-    if (data == NULL) {
+    type = g_type_from_name(info->class);
+    if (type == G_TYPE_INVALID) {
 	char buf[50];
 	g_warning("unknown widget class '%s'", info->class);
 	g_snprintf(buf, 49, "[a %s]", info->class);
 	ret = gtk_label_new(buf);
 	gtk_widget_show(ret);
     } else {
-	if (!data->typecode && data->get_type_func)
-	    data->typecode = data->get_type_func();
-	g_assert(data->new);
-	g_assert(data->typecode);
-	ret = data->new(self, data->typecode, info);
+	ret = get_build_data(type)->new(self, type, info);
     }
     glade_xml_set_common_params(self, ret, info);
     return ret;
@@ -1465,17 +1854,15 @@ void
 glade_xml_handle_internal_child(GladeXML *self, GtkWidget *parent,
 				GladeChildInfo *child_info)
 {
-    GladeWidgetBuildData *parent_build_data = NULL;
+    const GladeWidgetBuildData *parent_build_data = NULL;
     GtkWidget *child;
 
     /* walk up the widget heirachy until we find a parent with a
      * find_internal_child handler */
     while (parent_build_data == NULL && parent != NULL) {
-	parent_build_data = g_hash_table_lookup(widget_table,
-						G_OBJECT_TYPE_NAME(parent));
+	parent_build_data = get_build_data(G_OBJECT_TYPE(parent));
 
-	if (parent_build_data != NULL &&
-	    parent_build_data->find_internal_child != NULL)
+	if (parent_build_data->find_internal_child != NULL)
 	    break;
 
 	parent_build_data = NULL; /* set to NULL if no find_internal_child */
@@ -1534,16 +1921,16 @@ glade_xml_set_common_params(GladeXML *self, GtkWidget *widget,
 			    GladeWidgetInfo *info)
 {
     GList *tmp;
-    GladeWidgetBuildData *data;
+    const GladeWidgetBuildData *data;
 
     /* get the build data */
-    if (!widget_table)
-	widget_table = g_hash_table_new(g_str_hash, g_str_equal);
-    data = g_hash_table_lookup(widget_table, info->class);
+    data = get_build_data(G_OBJECT_TYPE(widget));
     glade_xml_add_signals(self, widget, info);
     glade_xml_add_accels(self, widget, info);
 
     gtk_widget_set_name(widget, info->name);
+    glade_xml_add_accessibility_info(self, widget, info);
+
 #if 0
     if (info->tooltip) {
 	gtk_tooltips_set_tip(self->priv->tooltips,
@@ -1572,14 +1959,29 @@ glade_xml_set_common_params(GladeXML *self, GtkWidget *widget,
     while (tmp) {
 	GladeDeferredProperty *dprop = tmp->data;
 
-	if (!strcmp(info->name, dprop->prop_value)) {
+	if (!strcmp(info->name, dprop->target_name)) {
 	    tmp = tmp->next;
 	    self->priv->deferred_props =
 		g_list_remove(self->priv->deferred_props, dprop);
 
-	    g_object_set(G_OBJECT(dprop->target),
-			 dprop->prop_name, G_OBJECT(widget),
-			 NULL);
+	    switch (dprop->type) {
+	    case DEFERRED_PROP:
+		g_object_set(G_OBJECT(dprop->d.prop.object),
+			     dprop->d.prop.prop_name, G_OBJECT(widget),
+			     NULL);
+		break;
+	    case DEFERRED_REL:
+		{
+		    AtkObject *target = gtk_widget_get_accessible(widget);
+
+		    add_relation(dprop->d.rel.relation_set,
+				 dprop->d.rel.relation_type, target);
+		    g_object_unref(dprop->d.rel.relation_set);
+		}
+		break;
+	    default:
+		g_warning("unknown deferred property type");
+	    }
 	    g_free(dprop);
 	} else {
 	    tmp = tmp->next;
@@ -1589,3 +1991,21 @@ glade_xml_set_common_params(GladeXML *self, GtkWidget *widget,
     if (data && data->build_children && info->children)
 	data->build_children(self, widget, info);
 }
+
+gchar **
+glade_xml_get_toplevel_names (GladeXML *self)
+{
+    int i;
+    gchar **ret;
+    GladeInterface *tree;
+    
+    tree = self->priv->tree;
+
+    ret = g_new (gchar *, tree->n_toplevels + 1);
+
+    for (i = 0; i < tree->n_toplevels; i++)
+	ret [i] = g_strdup (tree->toplevels [i]->name);
+    ret[i] = NULL;
+
+    return ret;
+}
